from typing import (
    Any,
    Dict,
    List,
    Optional,
)

from typing_extensions import (
    Literal,
    Protocol,
)

Impact = Literal["minor", "moderate", "serious", "critical"]

# Axe tests we want to actively pass in Galaxy, the next list controls tests.
BASELINE_AXE_PASSING_IDS = [
    "aria-roles",
]

# check all violations of this level as a baseline 'impact'...
BASELINE_VIOLATION_FILTER: Impact = "critical"
FORMS_VIOLATIONS = ["duplicate-id-aria"]  # https://github.com/galaxyproject/galaxy/issues/16188
# unless they are in this list...
KNOWN_VIOLATIONS = FORMS_VIOLATIONS + [
    "aria-required-attr",
    "aria-required-children",
    "aria-required-parent",
    "image-alt",  # test_workflow_editor.py::TestWorkflowEditor::test_existing_connections
    "label",
    "button-name",
    "select-name",
]
# Over time we hope known violations grows smaller until the violation
# filter can be lowered. Next level would be "serious".
# xref https://github.com/galaxyproject/galaxy/issues/16185


class AxeResult:
    def __init__(self, json: Dict[str, Any]):
        self._json = json

    @property
    def id(self) -> str:
        return self._json["id"]

    @property
    def description(self) -> str:
        return self._json["description"]

    @property
    def impact(self) -> Optional[Impact]:
        return self._json["impact"]

    @property
    def nodes(self):
        return self._json["nodes"]

    def is_impact_at_least(self, impact: Impact) -> bool:
        range_of_impacts = []
        if impact == "minor":
            range_of_impacts = ["minor", "moderate", "serious", "critical"]
        elif impact == "moderate":
            range_of_impacts = ["moderate", "serious", "critical"]
        elif impact == "serious":
            range_of_impacts = ["serious", "critical"]
        elif impact == "critical":
            range_of_impacts = ["critical"]
        return self.impact in range_of_impacts


class Violation(AxeResult):
    @property
    def message(self) -> str:
        nodes = self.nodes
        nodes_str = ", ".join([n["html"] for n in nodes])
        return f"AXE accessibility test violation found [{self.id}] with impact {self.impact}: {self.description}. Problem found in {nodes_str}."


class AxeResults(Protocol):
    def assert_passes(self, id: str) -> None:
        """"""

    def assert_does_not_violate(self, id: str) -> None:
        """"""

    def violations(self) -> List[Violation]:
        """"""

    # these next two could be refactored into a mixin...
    def violations_with_impact_of_at_least(self, impact: Impact) -> List[Violation]:
        """"""

    def assert_no_violations_with_impact_of_at_least(
        self, impact: Impact, excludes: Optional[List[str]] = None
    ) -> None:
        """"""


class RealAxeResults(AxeResults):
    def __init__(self, json: Dict[str, Any]):
        self._json = json

    def assert_passes(self, id: str) -> None:
        passing_results = self._json["passes"]
        result = _check_list_for_id(passing_results, id)
        assert result

    def assert_does_not_violate(self, id: str) -> None:
        violations = self._json["violations"]
        if result := _check_list_for_id(violations, id):
            violation = Violation(result)
            raise AssertionError(violation.message)

    def violations(self) -> List[Violation]:
        violations = self._json["violations"]
        return [Violation(v) for v in violations]

    def violations_with_impact_of_at_least(self, impact: Impact) -> List[Violation]:
        return [v for v in self.violations() if v.is_impact_at_least(impact)]

    def assert_no_violations_with_impact_of_at_least(
        self, impact: Impact, excludes: Optional[List[str]] = None
    ) -> None:
        excludes = excludes or []
        violations = self.violations_with_impact_of_at_least(impact)
        filtered_violations = [v for v in violations if v.id not in excludes]
        if filtered_violations:
            raise AssertionError(filtered_violations[0].message)


class NullAxeResults(AxeResults):
    """All assertions just pass because we're skipping Axe evaluation."""

    def assert_passes(self, id: str) -> None:
        pass

    def assert_does_not_violate(self, id: str) -> None:
        pass

    def violations(self) -> List[Violation]:
        return []

    # these next two could be refactored into a mixin...
    def violations_with_impact_of_at_least(self, impact: Impact) -> List[Violation]:
        return []

    def assert_no_violations_with_impact_of_at_least(
        self, impact: Impact, excludes: Optional[List[str]] = None
    ) -> None:
        pass


def assert_baseline_accessible(axe_results: AxeResults) -> None:
    for passing_id in BASELINE_AXE_PASSING_IDS:
        axe_results.assert_passes(passing_id)
    for violation in axe_results.violations_with_impact_of_at_least(BASELINE_VIOLATION_FILTER):
        violation_id = violation.id
        if violation_id not in KNOWN_VIOLATIONS:
            raise AssertionError(violation.message)


def _check_list_for_id(result_list: List[Dict[str, Any]], id) -> Optional[Dict[str, Any]]:
    for result in result_list:
        if result.get("id") == id:
            return result
    return None



import logging
import os
from typing import Union

try:
    from pyvirtualdisplay import Display
except ImportError:
    Display = None

from selenium import webdriver
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.safari.options import Options as SafariOptions

logger = logging.getLogger("selenium.webdriver.remote.remote_connection")
logger.setLevel(logging.WARNING)

DEFAULT_BROWSER = "auto"
DEFAULT_DOWNLOAD_PATH = "/tmp/"
LOGGING_PREFS = {
    "browser": "ALL",
}
DEFAULT_SELENIUM_BROWSER = "auto"
DEFAULT_SELENIUM_REMOTE = False
DEFAULT_SELENIUM_REMOTE_PORT = "4444"
DEFAULT_SELENIUM_REMOTE_HOST = "127.0.0.1"
DEFAULT_WINDOW_WIDTH = 1280
DEFAULT_WINDOW_HEIGHT = 1000
VALID_LOCAL_BROWSERS = ["CHROME", "FIREFOX"]
PYVIRTUALDISPLAY_UNAVAILABLE_MESSAGE = "pyvirtualdisplay must be installed to run this test configuration and is not, install with 'pip install pyvirtualdisplay'"


class ConfiguredDriver:
    driver: WebDriver

    def __init__(
        self,
        browser=DEFAULT_SELENIUM_BROWSER,
        remote=DEFAULT_SELENIUM_REMOTE,
        remote_host=DEFAULT_SELENIUM_REMOTE_HOST,
        remote_port=DEFAULT_SELENIUM_REMOTE_PORT,
        headless=False,
    ):
        self.config = dict(
            browser=browser,
            remote=remote,
            remote_host=remote_host,
            remote_port=remote_port,
            headless=headless,
        )
        if remote:
            driver = get_remote_driver(
                host=remote_host,
                port=remote_port,
                browser=browser,
            )
        else:
            driver = get_local_driver(
                browser=browser,
                headless=headless,
            )
        # Need larger than default window size for workflow index page, paired list creator.
        driver.set_window_size(DEFAULT_WINDOW_WIDTH, DEFAULT_WINDOW_HEIGHT)
        self.driver = driver

    def to_dict(self):
        return self.config

    @staticmethod
    def from_dict(as_dict):
        return ConfiguredDriver(**as_dict)


def get_local_browser(browser):
    if browser == "auto":
        if _which("chromedriver"):
            browser = "CHROME"
        elif _which("geckodriver"):
            browser = "FIREFOX"
        else:
            raise Exception("Selenium browser is 'auto' but neither geckodriver or chromedriver are found on PATH.")
    return browser


def get_local_driver(browser=DEFAULT_BROWSER, headless=False) -> WebDriver:
    browser = get_local_browser(browser)
    if browser not in VALID_LOCAL_BROWSERS:
        raise AssertionError(f"{browser} not in VALID_LOCAL_BROWSERS ({VALID_LOCAL_BROWSERS})")
    driver_to_class = {
        "CHROME": webdriver.Chrome,
        "FIREFOX": webdriver.Firefox,
    }
    driver_class = driver_to_class[browser]
    if browser == "CHROME":
        chrome_options = ChromeOptions()
        if headless:
            chrome_options.add_argument("--headless")
        prefs = {"download.default_directory": DEFAULT_DOWNLOAD_PATH}
        chrome_options.add_experimental_option("prefs", prefs)
        chrome_options.set_capability("goog:loggingPrefs", LOGGING_PREFS)
        return driver_class(options=chrome_options)
    else:
        firefox_options = webdriver.FirefoxOptions()
        firefox_options.set_preference("network.proxy.type", 2)
        firefox_options.set_preference("network.proxy.autoconfig_url", "http://127.0.0.1:9675")
        firefox_options.set_preference("browser.download.folderList", 2)
        firefox_options.set_preference("browser.download.dir", DEFAULT_DOWNLOAD_PATH)
        firefox_options.set_preference("browser.helperApps.neverAsk.saveToDisk", "application/octet-stream")
        return driver_class(options=firefox_options)


def get_remote_driver(host, port, browser=DEFAULT_BROWSER) -> WebDriver:
    # docker run -d -p 4444:4444 -v /dev/shm:/dev/shm selenium/standalone-chrome:3.0.1-aluminum
    options: Union[webdriver.ChromeOptions, webdriver.FirefoxOptions, webdriver.EdgeOptions, SafariOptions]
    if browser == "auto" or browser == "CHROME":
        options = webdriver.ChromeOptions()
        options.set_capability("goog:loggingPrefs", LOGGING_PREFS)
    elif browser == "FIREFOX":
        options = webdriver.FirefoxOptions()
    elif browser == "EDGE":
        options = webdriver.EdgeOptions()
    elif browser == "SAFARI":
        options = SafariOptions()
    else:
        raise Exception(f"Browser '{browser}' not supported.")
    executor = f"http://{host}:{port}/wd/hub"
    driver = webdriver.Remote(
        command_executor=executor,
        options=options,
    )
    return driver


def is_virtual_display_available():
    return Display is not None


def virtual_display_if_enabled(enabled):
    if enabled:
        if Display is None:
            raise Exception(PYVIRTUALDISPLAY_UNAVAILABLE_MESSAGE)
        display = Display(visible=0, size=(800, 600))
        display.start()
        return display
    else:
        return NoopDisplay()


class NoopDisplay:
    def stop(self):
        """No-op stop for consistent use with pyvirtualdisplay Display class."""


# Purposely copied from galaxy.util - just use galaxy.util if we decide
# galaxy_selenium should definitely depend on Galaxy/galaxy-lib.
def _which(file):
    # http://stackoverflow.com/questions/5226958/which-equivalent-function-in-python
    for path in os.environ["PATH"].split(":"):
        if os.path.exists(f"{path}/{file}"):
            return f"{path}/{file}"

    return None


__all__ = (
    "ConfiguredDriver",
    "get_local_driver",
    "get_remote_driver",
    "is_virtual_display_available",
    "virtual_display_if_enabled",
)
